#!/usr/bin/perl

#  Copyright 2002, 2003, 2004, 2005 University of Southern California
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation or under the same terms as perl itself.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package schedtop;

# Configurable defaults
# Don't change the defaults here, instead make yourself a config file, see %SCRIPTNAME%(1)
our $columns      	= 30;      # columns in grid
our $autocolumns    = 1;       # 1 or 0, set $columns based on term width
our $sleeptime    	= 20;      # seconds between refreshes
our $colorize     	= 1;       # 1 or 0
our $show_summary 	= 1;       # 1 or 0
our $compact_summary= 0;       # 1 or 0
our $show_grid    	= 1;       # 1 or 0
our $show_queue   	= 1;       # 1 or 0
our $show_qqueue  	= 1;       # 1 or 0
our $show_jobs    	= 1;       # 1 or 0
our $show_user    	= "all";   # show only a given user's jobs
our $show_onlyq    	= "";      # show only this queue
our @show_cpu     	= ();      # list of cpu numbers
our @host           = ();      # leave empty for localhost
our $maxrows        = 1500;    # maximum number of rows
our $maxcolumns     = 300;     # maximum number of columns
our $maxnodegrid	= -1;	   # max. number of CPUs on a node before it gets its own grid.
our $showncpus      = 0;       # display the ncpus job attribute (if using perl-PBS or Slurm)
our $fillbg         = 0;       # 1 or 0, fill the term background with black?
our $nodesort		= "mixed"; # mixed, mixed2, integer, lexical, or ordered
our $nospace        = 0;       # leave out spaces between the columns
our $compress_arrays= 1;       # 1 or 0
our $reletterize    = 0;

our %Jobs;

our $script_name;

#########################################################
### Nothing else to adjust below here

use strict;
use vars qw/$VERSION/;
use Curses;
use POSIX qw/strerror ceil/;
use File::Basename;

$VERSION = "5.02";

# init a few global vars
our %Job_of_letter;
our @Colors = ();
our $masterletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
our $letters = $masterletters;
our $underline = 0;
our %searchobject=();
our @nodesort_hostconf=();  # filled in by local rc config
our %nodesort_host	= ();      # filled in from @nodesort_hostconf and @host
our ( $y, $x, $Y, $X, $py, $px, $ly, $lx, $subY, $subX ) = ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
our ($CTRL_B, $CTRL_F, $CTRL_L, $CTRL_G, $CTRL_H);
our $pr;
our $pad;
our $cmdwin;
our $subwin;
our ($get_version, $get_server_data, $get_job_data, $set_node_state, $set_job_state);

# Command-line argument processing
# We pass in the scriptname (currently pbstop or slurmtop), the server name array,
# and the command line argument array.
sub process_args {
    $script_name = shift;
    $host[0] = shift;
    my @argv = @_;

    my @argvhosts=();
    while ( my $arg = shift @argv ) {
        if ( $arg eq '-s' ) {
            $sleeptime = shift @argv;
            $sleeptime =~ /^\d+$/ or $argv[0] = '-h';
            $sleeptime > 0 or $argv[0] = '-h';
        }
        elsif ( $arg eq '-c' ) {
            $columns = shift @argv;
            $columns =~ /^\d+$/ or $argv[0] = '-h';
            if ($columns == 0) {
               $autocolumns=1;
            } elsif ($columns > 0) {
               $autocolumns=0;
            } else {
               $argv[0] = '-h';
            }
        }
        elsif ( $arg eq '-m' ) {
            $maxnodegrid = shift @argv;
            $maxnodegrid =~ /^\d+$/ or $argv[0] = '-h';
            $maxnodegrid > 0 or $argv[0] = '-h';
        }
        elsif ( $arg =~ /^-(\d+)$/ ) {
            @show_cpu = split ( //, $1 );
        }
        elsif ( $arg eq '-u' ) {
            if (defined $argv[0] and $argv[0] =~ /^([^-]+)/) {
                $show_user=join(' ', split(',', shift @argv));
            } else {
                $argv[0] = '-h';
            }
        }
        elsif ( $arg eq '-q' ) {
            $show_onlyq = shift @argv;
        }
        elsif ( $arg eq '-C' ) {
            $colorize = !$colorize;
        }
        elsif ( $arg eq '-S' ) {
            $show_summary = !$show_summary;
        }
        elsif ( $arg eq '-G' ) {
            $show_grid = !$show_grid;
        }
        elsif ( $arg eq '-J' ) {
            $show_jobs = !$show_jobs;
        }
        elsif ( $arg eq '-Q' ) {
            $show_queue = !$show_queue;
        }
        elsif ( $arg eq '-t' ) {
            $show_qqueue = !$show_qqueue;
        }
        elsif ( $arg eq '-A' ) {
            $compress_arrays = !$compress_arrays;
        }
        elsif ( $arg eq '-n' ) {
            $nospace = 1;
        }
        elsif ( $arg eq '-fillbg' ) {
            $fillbg=1;
        }
        elsif ( $arg eq '-V' ) {
            print "$script_name $VERSION\n";
            # Copyright and authorship only applies to pbstop.
            if ($script_name eq "pbstop") {
                print "Copyright 2002, 2003, 2004 University of Southern California\n";
            }
            exit(0);
        }
        elsif ( $arg =~ /^@(.*)/ ) {
            push(@argvhosts, $1);
        }
        else {
            print "Usage:  $script_name [-s seconds] [-c columns] [-m maxcpus] [options] [\@host ...]\n";
            print "   Version: $VERSION\n";
            # Copyright and authorship only applies to pbstop.
            if ($script_name eq "pbstop") {
                print "   Copyright 2002, 2003, 2004 University of Southern California\n";
            }
            print "   \n";
            print "   -h  print this help menu and exit\n";
            print "   -V  print version and exit\n";
            print "   \n";
            print "   -A  toggle array job compression\n";
            print "   -c  node columns to display in the primary grid (0 = autoset)\n";
            print "   -C  toggle colorization\n";
            print "   -fillbg  fill background with black\n";
            print "   -G  toggle grid display\n";
            print "   -J  toggle jobs in grid display\n";
            print "   -m  max number of cpus in a node before it gets its own grid\n";
            print "   -n  no space between columns (compact display)\n";
            print "   -q  limit job view to specific queue\n";
            print "   -Q  toggle queue display\n";
            print "   -s  seconds between refreshes\n";
            print "   -S  toggle state summary display\n";
            print "   -t  toggle showing queued jobs (status = Q) in job display\n";
            print "   -u  limit job view to specific user(s)\n";
            exit(1);
        }
    }
    if(scalar @argvhosts > 0) {
        @host=@argvhosts;
    }
    undef @argvhosts;

    # setup %nodesort_host
    init_nodesort( @host );

    if ($show_user eq "all") {
       $show_user=0;
    } elsif ($show_user =~ /\bme\b/) {
       $show_user=~ s/\bme\b/$ENV{USER}/;
    }
    use vars qw/$SIGWINCH/;
    $SIGWINCH=0;
    $SIG{'WINCH'} = sub {$SIGWINCH=1; };
    $SIG{'INT'} = sub { endwin; exit(0); };
    $SIG{'TERM'} = sub { endwin; exit(0); };
    $SIG{'__DIE__'} = sub { endwin; die @_; };

    if (! -t STDOUT) {
       $colorize=0;
    }

    # Is this portable?
    $CTRL_B=chr(ord("B")-ord("@"));
    $CTRL_F=chr(ord("F")-ord("@"));
    $CTRL_L=chr(ord("L")-ord("@"));
    $CTRL_G=chr(ord("G")-ord("@"));
    $CTRL_H=chr(ord("H")-ord("@"));

    initscr;
    cbreak;
    noecho;
    getmaxyx( $Y, $X );

    $colorize=$colorize && has_colors();
    $pr = 0;
    if (has_colors()) {
       start_color;
       init_pairs();
    }
    $pad = newpad( $maxrows, $maxcolumns );
    $cmdwin = newwin( 1, $X - 1, $Y - 1, 0 );
    keypad( $cmdwin, 1 );
    $subwin=0;

    addstr $cmdwin, 0, 0, "Starting...";
    clrtoeol($cmdwin);
    refresh($cmdwin);
    move( $cmdwin, 0, 0 );
    clrtoeol($cmdwin);
}

###############################################################
## All subroutines below here
###############################################################

# main_loop() will 1) gather all of our data from the resource manager, 2) prep it a bit
# in letterize() and colorize(), 3) call update_display to draw our pretty
# grids and stuff, and finally calls 4) top_sleep which is where we spend most
# of our time.

# 1) We gather node and job data by calling the subroutine reference $job_info. It
# connects directly to the server and requests the desired data. Since this data 
# is kept between cycles around the main loop, we take some care to remove old data.
# The result is two large structures, one for Jobs and one for Nodes, which will 
# carry us through the rest of the entire program.

# 2) letterize() and colorize() are fairly unexciting, but they do assign
# letters and colors to each running job.  This info is stored in the large Job
# structure.  letterize() has probably the only original code left.

# 3) update_display(), by itself, is pretty boring.  It calls the functions
# responsible for the summary, colorful grid, and the job listing at the
# bottom.  show_grid() is pretty exciting; it first finds every node over
# $maxnodegrid, calls them "timesharing" and shoves them aside, draws a big
# colorful grid for what's left, and draws another colorful grid for the
# timesharing nodes.

# 4) top_sleep() is a big giant mess.  It is far too monolithic.  If anyone
# wants to chop it up a bit, feel free to send me patches!  Anyways, it loops
# around on user input until the time expires and it is time to return back up
# to main_loop().  In the meantime, it does everything the user requests,
# including grinding through the main data structures looking for stuff.  All of
# the code responsible for moving around the giant pad is here.


sub main_loop {
    my ($get_info, $get_version_local, $get_server_data_local, $get_job_data_local, 
            $set_node_state_local, $set_job_state_local) = @_;
    $get_version = $get_version_local;
    $get_server_data = $get_server_data_local;
    $get_job_data = $get_job_data_local;
    $set_node_state = $set_node_state_local;
    $set_job_state = $set_job_state_local;

    my $maxlen;
    my %Nodes;
    my %State_count;

    # temp vars
    my $node;

    # Main event loop.
    while (1) {

        %State_count	      = ();
        $State_count{_nodes}  = 0;
        $State_count{_anodes} = 0;
        $State_count{_procs}  = 0;
        $State_count{_aprocs} = 0;
        $State_count{_mprocs} = 0;
        $State_count{_rjobs}  = 0;
        $State_count{_njobs}  = 0;

        foreach my $server (@host) {
            my $warn = $get_info->($server, \%Nodes, \%Jobs, \%State_count);
            if ($warn) {
                printwarning($warn);
                sleep 3;
            }
            # Trim out old nodes that are no longer seen
            foreach my $node (keys %{$Nodes{$server}}) {
               if ($Nodes{$server}{$node}{seen} != 1) {
                   delete $Nodes{$server}{$node};
               } else {
                   $Nodes{$server}{$node}{seen}=0;
               }
            }
        }
        $maxlen = getmaxkeylen( \%Nodes );

        # If not set via config file or command line argument, 
        # set @show_cpu to be the max number of cores per node.
        # If that number is bigger than will fit on one line, or
        # bigger than $maxnodegrid, then recompute $State_count{_mprocs} 
        # so it excludes any such nodes. A side-effect of this is that, 
        # unless $maxnodegrid is explicitly specified, nodes that don't fit on
        # one line will be automatically displayed in the timeshare grid (@ts).
        if (scalar @show_cpu == 0) {
            my $linefit = int(($X - $maxlen - 4) / (1 - $nospace + $State_count{_mprocs}));
            if ($linefit == 0 || ($maxnodegrid != -1 && $State_count{_mprocs} > $maxnodegrid)) {
                $State_count{_mprocs} = 0;
                foreach my $server (@host) {
                    foreach my $node (keys %{$Nodes{$server}}) {
                        my $np = $Nodes{$server}{$node}{np};
                        $linefit = int(($X - $maxlen - 4) / (1 - $nospace + $np));
                        if ($linefit > 0 && ($maxnodegrid != -1 && $np <= $maxnodegrid)) {
                            $State_count{_mprocs} = $State_count{_mprocs} > $np
                                    ? $State_count{_mprocs} : $np;
                        }
                    }
                }
            }
            @show_cpu = (0 .. $State_count{_mprocs}-1);
        }
        # If not set via config file or command line argument, 
        # set $maxnodegrid to be the number of CPUs to be shown.
        if ($maxnodegrid == -1) {
            $maxnodegrid = scalar @show_cpu;
        }
        # trim out old jobs that are no longer seen
        foreach my $job (keys %Jobs) {
            if (!exists $Jobs{$job}{seen} or !defined $Jobs{$job}{seen} or $Jobs{$job}{seen} != 1) {
               delete $Jobs{$job};
            } else {
               $Jobs{$job}{seen}=0;
            }
            if (! $showncpus) {
                if (exists $Jobs{$job}{ncpus} or exists $Jobs{$job}{nodes}) {
                    $showncpus = 1;
                }
            }
        }
        letterize( \%Jobs );
        colorize( \%Jobs );

        update_display( \%State_count, \%Nodes, \%Jobs, $maxlen,
            $State_count{"_mprocs"} );
        -t STDOUT or do { endwin; exit; };

        my $status = top_sleep( \%State_count, \%Nodes, \%Jobs, $maxlen,
            $State_count{"_mprocs"} );
        if ($status == 1) {
            return 0;
        }
    }
}

sub update_display {
    my $foo;
    move( $pad, 0, 0 );
    getmaxyx( $Y, $X );

    $y = 0, $x = 0;

    printwarning("Displaying summary...") if $^W;
    $show_summary and show_state_summary( $_[0] );

    printwarning("Displaying grid...") if $^W;
    $show_grid and show_grid( $_[2], $_[1], $_[3], $_[4] );

    printwarning("Displaying queue...") if $^W;
    $show_queue and show_queue( $_[2], $_[0] );

    printwarning("done.") if $^W;

    getyx( $pad, $ly, $foo );
    clrtobot($pad);

    pnoutrefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
    mvwin( $cmdwin, $Y - 1, 0 );
    update_subwin(@_);
    doupdate();
}

sub show_state_summary {
    my $t  = 1;
    my $t2 = 1;

    addstr $pad,
      sprintf(
        "Usage Totals: %d/%d %s, %d/%d %s, %d/%d %s",
        ${ $_[0] }{_aprocs},
        ${ $_[0] }{_procs},
        "Procs",
        ${ $_[0] }{_anodes},
        ${ $_[0] }{_nodes},
        "Nodes",
        ${ $_[0] }{_rjobs},
        ${ $_[0] }{_njobs},
        "Jobs Running"
      );

    my ( $y1, $x1 );
    getyx( $pad, $y1, $x1 );
    addstr $pad, " " x ( $X - $x1 - 8 );

    # Asbed asked for the time
    addstr $pad, sprintf( "%02d:%02d:%02d", ( localtime() )[ 2, 1, 0 ] );

    my $line;
    my @states = sort grep !/^_/, keys %{ $_[0] };

    if ( $compact_summary ) {
        move( $pad, ++$y, $x = 0 );
        addstr( $pad, 1, 0, "Node States: " );
        for ( my $i = 0 ; defined $states[$i] ; $i++ ) {
            $line= " ".${ $_[0] }{ $states[$i] }." ". $states[$i];
            $line.= defined $states[$i+1] ? "," : "";
            getyx( $pad, $y1, $x1 );
            if ( $X-$x1-1 < length($line) ) {
                clrtoeol($pad);
                move( $pad, ++$y, $x = 0 );
                addstr ($pad, " " x 13);
            }
            addstr ($pad, $line);
        }
        move( $pad, ++$y, $x = 0 );
            
    } else {
        my $statecolumns=int(($X-14)/25);
        for ( my $i = 0 ; defined $states[$i] ; ) {
            move( $pad, ++$y, $x = 0 );
            $line = " " x 14;

            for (my $k=1; $k<=$statecolumns; $k++) {
               last unless (defined $states[$i]);
               $line.= sprintf( "%4s %-20s", ${ $_[0] }{ $states[$i] }, $states[$i] );
               $i++;
            }
            $line .= " " x ( $X - ( length($line) + 14 ) );
            clrtoeol($pad);
            addstr $pad, $line;
        }
        addstr( $pad, 1, 0, "Node States:" );
        move( $pad, ++$y, $x = 0 );
    }
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );
    clrtoeol($pad);
}

sub show_grid {
    my ( $jobs, $allnodes, $maxlen, $maxprocs ) = @_;
    my ( $foo, $tmpx );
    $lx = 0;

    # We now assume that this condition occurs when no node has a number 
    # of cpus small enough to be displayed on one line (e.g., SGI UV).
    # So we are commenting out this logic.
    #if ( !scalar @show_cpu ) {
    #    addstr $pad, "  No CPUs selected!";
    #    clrtoeol($pad);
    #    return;
    #}
    foreach my $server (keys %$allnodes ) {
        my (@cluster, @ts);
        foreach my $node ( node_sort_func($server, $allnodes->{$server}) ) {
            $$allnodes{$server}{$node}{np} > $maxnodegrid
                ? push(@ts, $node)
                : push(@cluster, $node);
        }
        # loop through each node, in lines and columns
        my $col  = 0;
        
        my ($this_cpu, $jobid, $letter, $color, $underline, $nodep, $jobp);
        if (defined $cluster[0]) {
            printvcpuline($maxlen);
            if ($autocolumns) {
                $columns = int(($X - $maxlen - 4) / (1 - $nospace + scalar @show_cpu));
                $columns = $columns < scalar @cluster ? $columns : scalar @cluster;
            }
            my $headerspaces = scalar @show_cpu - $nospace;
            printnumberline( $maxlen, $headerspaces, $columns );
            printdashline( $maxlen, $headerspaces, $columns );

            my $gridrows=0;
            foreach my $node ( @cluster ) {
                $col = 0 if $col >= $columns;
                $nodep=$allnodes->{$server}{$node};
                (addstr $pad, sprintf "  %${maxlen}s ", $node) if $col == 0;
                addstr $pad, "  " if ( $col != 0 and $col % 10 == 0 );
        
                foreach $this_cpu (@show_cpu) {
                    if ( $this_cpu > $nodep->{np} - 1 ) {
                        addstr $pad, " ";
                    } else {
                        if (exists $nodep->{job}{$this_cpu}) {
                            $jobid = $nodep->{job}{$this_cpu};
                            if (exists $jobs->{$jobid}) {
                                $jobp=$jobs->{$jobid};
                                if (exists $jobp->{letter}) {
                                    if (($show_user ? $show_user =~ /\b$jobp->{user}\b/ : 1)
                                    and ($show_onlyq ? $show_onlyq =~ /\b$jobp->{queue}\b/ : 1)) {
                                        $letter= $jobp->{letter};
                                        $color = $jobp->{color};
                                        $underline = $jobp->{underline};
                                        printcpustate($jobid, $letter, $nodep->{state}, $color, $underline);
                                    } else {
                                        my $colorize_save = $colorize;
                                        $colorize = 0;
                                        printcpustate($jobid, '@', $nodep->{state}, 0, 0);
                                        $colorize = $colorize_save;
                                    }
                                } else {
                                    printcpustate(0,0,$nodep->{state},0,0);
                                }
                            } else {
                                # The job was deleted in between the time we captured
                                # the node info and the job info from pbs/slurm.
                                printcpustate(0,0,$nodep->{state},0,0);
                            }
                        } else {
                            printcpustate(0,0,$nodep->{state},0,0);
                        }
                    }
                }
                addstr $pad, " " unless $nospace;
                $col++;
                getyx( $pad, $foo, $tmpx );
                $lx = $lx > $tmpx ? $lx : $tmpx;

                clrtoeol($pad);
                if ($col >= $columns) {
                  move( $pad, ++$y, $x = 0 );
                  $gridrows++;
                }
                if ($gridrows >= (($Y>10?$Y:10)-6-(2*$show_summary))) {
                   $gridrows=0;
                   clrtoeol($pad);
                   move( $pad, ++$y, $x = 0 );
                   printnumberline ( $maxlen, $headerspaces, $columns );
                }
            }
            clrtoeol($pad);
            move( $pad, ++$y, $x = 0 ) if $col != $columns;
            printdashline( $maxlen, $headerspaces, $columns );
        }
        # Display @ts cpus in groups of 10.
        my $cpu_group = 10;
        my $max_columns = int(($X - $maxlen - 4) / (1 - $nospace + $cpu_group));
        # If no primary cluster grid, use the # columns of the timeshare grid as
        # $columns. Only used to display the number of autoset columns in the help screen.
        if (scalar @cluster == 0) {
            $autocolumns = 1;
            $columns = $max_columns;
        }
        my $first_ts = 1;
        foreach my $node (@ts) {
            $nodep=$allnodes->{$server}{$node};
            my $ts_columns;
            $ts_columns = int(($nodep->{np} + $cpu_group - 1) / 10);
            $ts_columns = $ts_columns < $max_columns ? $ts_columns : $max_columns;
            if (!($first_ts && scalar @cluster == 0)) {
                clrtoeol($pad);
                move( $pad, ++$y, $x = 0 );
            }
            $first_ts = 0;
            my $headerspaces = $cpu_group - $nospace;
            printnumberline( $maxlen, $headerspaces, $ts_columns );
            printdashline( $maxlen, $headerspaces, $ts_columns );
        
            my $col  = 0;
            foreach $this_cpu (0 .. $nodep->{np}-1) {
                $col = 0 if $col >= $ts_columns;
                if ($this_cpu == 0) {
                    addstr $pad, sprintf "  %${maxlen}s ", $node;
                } elsif ($this_cpu != 0 and $this_cpu % $cpu_group == 0) {
                    addstr $pad, " " unless $nospace;
                    $col++;
                    if ($col >= $ts_columns) {
                        clrtoeol($pad);
                        move( $pad, ++$y, $x = 0 );
                        addstr $pad, " " x (2 + $maxlen + 1);
                    } elsif ($col != 0 and $col % 10 == 0) {
                        addstr $pad, "  ";
                    }
                }
                if (exists $nodep->{job}{$this_cpu}) {
                    $jobid = $nodep->{job}{$this_cpu};
                    if (exists $jobs->{$jobid}) {
                        $jobp=$jobs->{$jobid};
                        if (exists $jobp->{letter}) {
                            if (($show_user ? $show_user =~ /\b$jobp->{user}\b/ : 1)
                            and ($show_onlyq ? $show_onlyq =~ /\b$jobp->{queue}\b/ : 1)) {
                                $letter= $jobp->{letter};
                                $color = $jobp->{color};
                                $underline = $jobp->{underline};
                                printcpustate($jobid, $letter, $nodep->{state}, $color, $underline);
                            } else {
                                my $colorize_save = $colorize;
                                $colorize = 0;
                                printcpustate($jobid, '@', $nodep->{state}, 0, 0);
                                $colorize = $colorize_save;
                            }
                        } else {
                            printcpustate(0,0,$nodep->{state},0,0);
                        }
                    } else {
                        # The job was deleted in between the time we captured
                        # the node info and the job info from pbs/slurm.
                        printcpustate(0,0,$nodep->{state},0,0);
                    }
                } else {
                    printcpustate(0,0,$nodep->{state},0,0);
                }
            }
            getyx( $pad, $foo, $tmpx );
            $lx = $lx > $tmpx ? $lx : $tmpx;

            clrtoeol($pad);
            move( $pad, ++$y, $x = 0 ) if $col != $ts_columns;
            printdashline( $maxlen, $headerspaces, $ts_columns );
        }
    }
    clrtoeol($pad);
    addstr( $pad, $y, 0, " [.] idle  [@] busy  [*] down  [%] offline  [!] other  [?] unknown" );
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );

}

sub cmp_jobarrayindex {
    $a = shift;
    $b = shift;
    my ($a_pre, $a_suf) = $a =~ /(\d+)(?:\[(\d+)[\d,-]*\])?/;
    my ($b_pre, $b_suf) = $b =~ /(\d+)(?:\[(\d+)[\d,-]*\])?/;

    if (defined $a_suf && defined $b_suf) {
        if ($a_pre == $b_pre) {
            return $a_suf <=> $b_suf;
        }
    }
    return $a_pre <=> $b_pre;
}

# Print out the job queue
sub show_queue {
    my ($jobs, $state_count) = @_;

    attron( $pad, A_BOLD );
    if ($showncpus) {
      addstr( $pad,"      Job#         Username Queue        Jobname          CPUs/Nodes S   Elapsed/Requested");
    } else {
      addstr( $pad,"      Job#         Username Queue        Jobname          Nodes S   Elapsed/Requested");
    }
    attroff( $pad, A_BOLD );
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );

    # Note: we never print out the server name, it is expected that the user will
    # recognize jobs by their jobid or queue name.

    my %seen_jobarrays = ();

    my ($jobptr,$state);
    my $prevArray = 0;
    foreach my $job (
            # Sort first by server, then by queue, last by jobid
            sort {
                (defined $jobs->{$a}{server} 
                    && defined $jobs->{$b}{server} 
                    && $jobs->{$a}{server} cmp $jobs->{$b}{server}) or 
                (defined $jobs->{$a}{queue} 
                    && defined $jobs->{$b}{queue} 
                    && $jobs->{$a}{queue} cmp $jobs->{$b}{queue}) or
                (cmp_jobarrayindex($a,$b))
            } keys %{$jobs} ) {
        $jobptr=$jobs->{$job};
        $state = $jobptr->{state};
        my $ncpus = 0;
        if ($showncpus) {
            $ncpus = exists $jobptr->{ncpus} ? $jobptr->{ncpus} : 
                    exists $jobptr->{nodes} ? count_nodestring($jobptr->{nodes}) : 0;
        }
        my $nnodes = exists $jobptr->{ncount} ? $jobptr->{ncount} : 0;
        my $elpt = exists $jobptr->{elpt} ? $jobptr->{elpt} : "";
        my $reqt = exists $jobptr->{reqt} ? $jobptr->{reqt} : "";

        if ( defined $jobptr->{user} ) {
            next unless $show_user ? $show_user =~ /\b$jobptr->{user}\b/ : 1;
            next unless $show_onlyq ? $show_onlyq =~ /\b$jobptr->{queue}\b/ : 1;

            # Ignore Q and H jobs if toggled off.
            # =~ /Q|H/ is slower
            if ( !( ($state eq "Q" or $state eq "H") and !$show_qqueue ) ) {
                my $ignore_array = 0;
                my @jobarray = $job =~ /(\d+)\[([\d,-]+)\]/;
                if ($compress_arrays == 1 && scalar @jobarray > 0) {
                    if ( exists $seen_jobarrays{$jobarray[0]} ) {
                        # just asserting here that we've already seen the lowest index.
                        #die "BAH not sorted" if $seen_jobarrays{$jobarray[0]} > $jobarray[1];

                        # Sum the number of cores for all array job members.
                        if ($showncpus and !($state =~ /Q|H|C/)) {
                            my $cloneref = \@{$seen_jobarrays{$jobarray[0]}};
                            $cloneref->[3] += $ncpus;
                            $prevArray = $jobarray[0];
                        }
                        next;
                    }
                    $ignore_array = 1;
                }
                # When finished with previous array job, append # of active cores for entire array.
                if ($prevArray) {
                    my $cloneref = \@{$seen_jobarrays{$prevArray}};
                    my $savedJobp = $jobs->{$cloneref->[2]};
                    if ( defined $savedJobp->{letter} ) {
                        # Reposition cursor to end of summary line.
                        my $savedY = $cloneref->[0];
                        my $savedX = $cloneref->[1];
                        move( $pad, $savedY, $savedX );
                        # Append number of cores represented by compressed array job letter in the grid.
                        my $totNcpus = $cloneref->[3];
                        addstr $pad, "  (Array ";
                        print_colored_letter($savedJobp->{letter}, $savedJobp->{color}, $savedJobp->{underline});
                        addstr $pad, " = $totNcpus active cores)";
                        move( $pad, $y, $x );
                    }
                    $prevArray = 0;
                }
                addstr $pad, "  ";
                if ( defined $jobptr->{letter} and !($state =~ /Q|H|C/)) {
                    print_colored_letter($jobptr->{letter}, $jobptr->{color}, $jobptr->{underline});
                    addstr $pad, " = ";
                }
                else {
                    addstr $pad, "    ";
                }
                # The job is exiting or has exited, grab an eyeball
                if ($state eq "E" or $state eq "C") {
                    attron($pad, A_REVERSE);
                }
                my $jobstr;
                if ($showncpus) {
                    $jobstr = sprintf("%-12s %-8.8s %-12.12s %-16.16s %4.4s/%-4.4s  %1s %9.9s/%-8.9s",
                            $job, $jobptr->{user}, $jobptr->{queue}, $jobptr->{jname},
                            $ncpus, $nnodes, $state, $elpt, $reqt);
                    addstr $pad, $jobstr;
                } else {
                    $jobstr = sprintf("%-12s %-8.8s %-12.12s %-16.16s %5.5s %1s %9.9s/%-8.9s",
                            $job, $jobptr->{user}, $jobptr->{queue}, $jobptr->{jname},
                            $nnodes, $state, $elpt, $reqt);
                    addstr $pad, $jobstr;
                }
                if ( scalar @jobarray > 0 ) {
                    @{$seen_jobarrays{$jobarray[0]}} = ( $y, 6+length($jobstr), 
                            $job, (!($state =~ /Q|H|C/) ? $ncpus : 0) );
                }
                if ($state eq "E" or $state eq "C") {
                   attroff($pad, A_REVERSE);
                }
                clrtoeol($pad);
                move( $pad, ++$y, $x = 0 );

                if ($ignore_array == 1) {
                    #addstr( $pad,"      $job in an array of jobs press \"A\" to see them all");
                    #clrtoeol($pad);
                    #move( $pad, ++$y, $x = 0 );
                }
            }
        }
    }
    # If last job is a compressed array job, append # of active cores for entire array.
    if ($prevArray) {
        my $cloneref = \@{$seen_jobarrays{$prevArray}};
        my $savedJobp = $jobs->{$cloneref->[2]};
        if ( defined $savedJobp->{letter} ) {
            # Reposition cursor to end of summary line.
            my $savedY = $cloneref->[0];
            my $savedX = $cloneref->[1];
            move( $pad, $savedY, $savedX );
            # Append number of cores represented by compressed array job letter in the grid.
            my $totNcpus = $cloneref->[3];
            addstr $pad, "  (Array ";
            print_colored_letter($savedJobp->{letter}, $savedJobp->{color}, $savedJobp->{underline});
            addstr $pad, " = $totNcpus active cores)";
            move( $pad, $y, $x );
        }
        $prevArray = 0;
    }
    clrtoeol($pad);

    # If we have more jobs than can fit in the defined number of rows for our pad, 
    # then print status bar warning for user to increase $maxrows in config file.
    if ($y > $maxrows) {
        printwarning("To see all jobs, set 'maxrows=" . 
            (ceil(($state_count->{_njobs} + 1) / 100) * 100) . 
            "' or more in ~/.${script_name}rc or /etc/${script_name}rc and restart $script_name");
        sleep 3;
    }
}

# Pass in a reference to all jobs, and we'll assign a letter to each one.
sub letterize {
    my $Jobs    = shift;

    # The original pbstop only used one letter per job, and this info was held
    # in  %Job_of_letter.  Now that we've thrown out that limitation, it is only
    # used to note the fact that _someone_ is using that letter.  If a job gets
    # a letter that is already assigned, the second one will be noted in 
    # %Job_of_letter, but that's ok because we don't care _who_ has that letter.

    # remove info about jobs already assigned a letter
    foreach my $l ( keys %Job_of_letter ) {
        delete $Job_of_letter{$l} if ( exists $Jobs->{ $Job_of_letter{$l} }{user} );
    }

    # declare a hash of array jobs
    my %array_jobs = ();

    # pick a letter if not already choosen
    foreach my $job ( keys %{$Jobs} ) {
        next if !defined $Jobs->{$job}{state};
        next if $Jobs->{$job}{state} =~ /Q|H/;
	
        # give array jobs letters when unlettered or changing compress modes
        my @jobarray = ();
        my $as_array = 0;
        if ( !exists $Jobs->{$job}{letter} || $reletterize == 1) {
            # do we want to treat it as part of a job array
            @jobarray = $job =~ /(\d+)\[\d+\]/;
            if ($compress_arrays == 1 && scalar @jobarray > 0) {
                $as_array = $jobarray[0];
            }
            # if 2nd-Nth compressed array job, give it the same letter as the 1st
            if ($as_array != 0 && exists $array_jobs{$as_array}) {
                my $job_to_clone = $array_jobs{$as_array};
                $Jobs->{$job}{letter} = $Jobs->{$job_to_clone}{letter};
                $Jobs->{$job}{underline} = $Jobs->{$job_to_clone}{underline};
                next;
            }
        }
        # give regular jobs letters only when unlettered
        if ( !exists $Jobs->{$job}{letter} || ($reletterize == 1 && scalar @jobarray > 0)) {
            # find a letter that isn't already taken
            my $user = $Jobs->{$job}{user};
            my $l = substr( $user, 0, 1 );
            if ( exists $Job_of_letter{$l} ) {
                $l = uc($l);
                if ( exists $Job_of_letter{$l} ) {
                    if (length $letters <= 0) {
                        # replenish our supply of letters
                        $colorize or printwarning("Reusing letters on B&W terminal.");
                        $letters = $masterletters;
                        $underline=!$underline;
                    }
                    $letters =~ s/(.)//;
                    $l = $1;
                }
            }
            $Job_of_letter{$l} = $job;
            $Jobs->{$job}{letter} = $l;
            $Jobs->{$job}{underline} = $underline;
            $letters =~ s/$l//;
            if ($as_array != 0) {
                # add it to my arrayjobs hash
                $array_jobs{$as_array}=$job;
            }
        }
    }
}

# Pass in a reference to all jobs, and we'll assign a color to each one.
# We do this regardless of $colorize, so that the info is available if 
# the user changes color preference.
sub colorize {
    my $Jobs = shift or return;

    # declare a hash of array jobs
    my %array_jobs;

    foreach my $job ( keys %{$Jobs} ) {
        next if defined $Jobs->{$job}{color};

        # do we want to treat it as part of a job array
        my $as_array = 0;
        my @jobarray = $job =~ /(\d+)\[\d+\]/;
        if (scalar @jobarray > 0) {
            $as_array = $jobarray[0];
        }
        if ( $as_array != 0 && exists $array_jobs{$as_array} ) {
            my $job_to_clone = $array_jobs{$as_array};
            $Jobs->{$job}{color} = $Jobs->{$job_to_clone}{color};
            next;
        }
        scalar @Colors == 0 and @Colors = init_colors();
        $Jobs->{$job}{color} = shift @Colors;
        if ($as_array != 0) {
            # add it to my arrayjobs hash
            $array_jobs{$as_array}=$job;
        }
    }
}

# This sucks, I wanted to separate printing from colors,
# but I can't just pass back color escape strings.
# I'm forced to combine them here.
sub print_colored_letter {
    my ($letter, $color, $underline) = @_;
    $colorize or do { addstr $pad, $letter; return };

    attron( $pad, A_BOLD | COLOR_PAIR( $color ) | ($underline && A_UNDERLINE) );
    addstr $pad, $letter;
    attroff( $pad, A_BOLD | COLOR_PAIR( $color ) | ($underline && A_UNDERLINE) );
}

# Used to find the longest hostname to align the left side of the node grid
sub getmaxkeylen {
    my $maxlen_ = 0;
    my $server;
    foreach $server ( keys %{ $_[0] } ) {
        foreach ( sort keys %{${ $_[0] }{$server}} ) {
            $maxlen_ =
              length($_) > $maxlen_
              ? length($_)
              : $maxlen_;
        }
    }
    return $maxlen_;
}

# This is each character inside the node grid.
sub printcpustate {
    my ($job, $letter, $state, $color, $underline) = @_;
    if ($job) {
        printwarning("$job has no letter") unless $letter;
        $letter='&' unless $letter;
    }

    #FIXME# I'm ignoring job-sharing here because I've never seen it used.
    #FIXME# It would be more correct to use the constants defined in
    #FIXME# pbs_ifl.h, but that would break compatibility with non-perl-PBS
    #FIXME# environs.
    if ( $job and $show_jobs ) {
       print_colored_letter( $letter, $color, $underline );
    }
    elsif ( $job and !$show_jobs ) {
       addch $pad, "@";
    }
    elsif ( $state =~ /down/ ) {
       print_colored_letter( "*", 1, 0 );   # 1 is red on black
    }
    # Down state takes precedence over offline state.
    elsif ( $state =~ /offline/ ) {
       addch $pad, "%";
    }
    elsif ( $state =~ /job-exclusive/ ) {
       addch $pad, '@';
    }
    elsif ( $state =~ /busy/ ) {
       addch $pad, "@";
    }
    elsif ( $state =~ /reserve/ ) {
       addch $pad, "@";
    }
    elsif ( $state =~ /unknown/i ) {
       print_colored_letter( "?", 1, 0 );
    }
    elsif ( $state =~ /free/ ) {
       addch $pad, ".";
    }
    else {
       print_colored_letter( "!", 1, 0 );
    }
}

# Print the list of visible CPUs above the node grid.
sub printvcpuline {
    my $maxlen=shift;
    # inform the user of the visible CPUs.
    if ( scalar @show_cpu == 1 ) {
        #addstr $pad, "  CPU $show_cpu[0]      ";
        clrtoeol($pad);
    }
    #elsif ( scalar @show_cpu == $maxprocs ) {
        #addstr "   " . " " x ($maxlen);
    #}
    else {
        addstr $pad, " " x ( $X - 1 );
        move( $pad, $y, 0 );
        addstr $pad, " Visible CPUs: " . join ( ",", @show_cpu );
        clrtoeol($pad);
        move( $pad, ++$y, $x = 0 );
    }
}

# Print the line of dashes above and below the node grid.
sub printdashline {
    my $maxlen = shift;
    my $spaces = shift;
    my $columns = shift;

    my $line = "   " . " " x $maxlen;
    for ( my $i = 0 ; $i < $columns ; $i++ ) {
        $line .= "--" if ( $i != 0 and $i % 10 == 0 );
        $line .= "-" . "-" x $spaces;
    }
    $line =~ s/-$// unless $nospace;    # oops, we printed one extra, erase it
    addstr $pad, $line;
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );
}

# Print the repetitive line of numbers along the top of the node grid.
sub printnumberline {
    my $maxlen = shift;
    my $spaces = shift;
    my $columns = shift;

    my $line= " " x ($maxlen+3); # = "   " . " " x $maxlen;
    for ( my $i = 0, my $j = 0 ; $i < $columns ; $i++, $j++ ) {
        if ( $i != 0 and $i % 10 == 0 ) {
            $line .= "  ";
            $j = 0;
        }
        $line .= ( ( $j + 1 ) % 10 ) . " " x $spaces;
    }
    addstr $pad, $y, 0, $line;
    clrtoeol($pad);
    move( $pad, ++$y, $x = 0 );

}

# This is used in top_sleep to annoy the user.
sub printwarning {
    attron( $cmdwin, A_REVERSE );
    addstr $cmdwin, 0, 0, join(" ", @_);
    attroff( $cmdwin, A_REVERSE );
    clrtoeol($cmdwin);
    refresh($cmdwin);
}

# This is used in top_sleep to solicit the user.
sub getstring {
    my $input="";
    my $ch;
    my $x=0;
    addstr $cmdwin, 0, 0, join(" ", @_);
    $x=length join(" ", @_);
    clrtoeol($cmdwin);
    refresh($cmdwin);
    echo;
    nodelay( $cmdwin, 0 );
    #getstr( $cmdwin,  $input );
    $ch=getch($cmdwin);
    while (1) {
        $ch=getch($cmdwin);
        if ($ch eq ERR) {  # ERR returned on timeout
            next;

        # why is this so freaking complicated??
        } elsif ($ch eq KEY_BACKSPACE or $ch eq KEY_DC or $ch eq $CTRL_H) {
            if (length $input) {
               $x--;
               move($cmdwin, 0, $x);
               delch($cmdwin);
               refresh($cmdwin);
               $input=~s/.$// ;
            } elsif ($ch eq KEY_BACKSPACE) {
               move($cmdwin, 0, $x);
            }
        } elsif ($ch eq $CTRL_G) {  # user abort
            $input="";
            last;
        } elsif ($ch eq "\n") {
            last;
        } else {
            $x++;
            $input.=$ch;
        }
    }
    $input =~ s/^\s+//;
    $input =~ s/\s+$//;
    noecho;
    move( $cmdwin, 0, 0 );
    clrtoeol($cmdwin);
    refresh($cmdwin);
    return $input;
}

   
sub print_serverstatus_window {
    my $server=shift;
    my @data;

    my $err = $get_server_data->($server, \@data);
    if ($err) {
        destroy_subwin();
        printwarning($err);
        return;
    }
    print_status_window("$server", \@data);
}

# Since we don't store enough info about jobs in %Jobs, go ahead and get
# it from the server
sub print_jobstatus_window {
    my $job=shift;
    my $server=shift;
    my @data;
    my $joberr;

    my $err = $get_job_data->($server, $job, \@data, \$joberr);
    if ($err) {
        destroy_subwin();
        printwarning($err);
        return;
    }
    if ($joberr) {
        print_status_window("$job.$server", \@data, $joberr);
    } else {
        print_status_window("$job.$server", \@data, "'l' for job load report, 'd'elete, 'H'old, 'R'elease, 'r'erun");
    }
}

sub print_jobloadstatus_window {
   my $job=shift;
   my $server=shift;
   my $allnodes=shift;

   my @loads;
   my $value;
   my $freephys;
   my $sessions;
   foreach $server ( sort keys %{ $allnodes } ) {
NODE:
      foreach my $node ( sort keys %{$allnodes->{$server}} ) {
         foreach my $this_cpu (0 .. $allnodes->{$server}{$node}{np}-1) {
             if (exists $allnodes->{$server}{$node}{job}{$this_cpu} and 
                 exists $allnodes->{$server}{$node}{status}{loadave}) {
                if ($job eq $allnodes->{$server}{$node}{job}{$this_cpu}) {
                    $value ="load: ".$allnodes->{$server}{$node}{status}{loadave};
                    {  # yes, I'm TOTALLY cheating here.  Sue me.
                       local $^W=0; 
                       $freephys=int(($allnodes->{$server}{$node}{status}{physmem} -
                             ($allnodes->{$server}{$node}{status}{totmem} -
                             $allnodes->{$server}{$node}{status}{availmem}))/1024);
                       $sessions=($allnodes->{$server}{$node}{status}{nsessions} =~ /^\?/)
                                      ? 0
                                      : $allnodes->{$server}{$node}{status}{nsessions};
                                      #: scalar (split / /,$allnodes->{$server}{$node}{status}{sessions});
                       $value.="  physmem: ".int($allnodes->{$server}{$node}{status}{physmem}/1024)."MB";
                       $value.=" avail: ${freephys}MB";
                       $value.="  sessions: $sessions";
                               
                    }
                    push(@loads, { name => $node,
                                   value =>  $value });
                    next NODE;
                }
             }
         }
      }
   }
                             
   if (scalar @loads < 1) {
        push(@loads, { name => "load report",
                       value => "The job is not running (i.e., it is queued or ended)"});
   }
   print_status_window("$job Load Report", \@loads, "'j'ob details");

}

# We already have everything we need to know about nodes in our big Nodes
# struct, so just pull info from there.
sub print_nodestatus_window {
   my $nodename=shift or return;
   my $ref=shift or return;
   
   my (@attrs, $name, $value);
   push(@attrs, { name => "state", value => $ref->{state} });
   push(@attrs, { name => "np", value => $ref->{np} });
   if (exists $ref->{properties}) {
      push(@attrs, { name => "properties", value => $ref->{properties} });
   }
   push(@attrs, { name => "ntype", value => $ref->{ntype} });
   while (($name, $value) = each %{$ref->{status}}) {
      push(@attrs, { name => "status.$name", value => $value });
   }
   if (exists $ref->{note}) {
     push(@attrs, { name => "note", value => $ref->{note} });
   }

   #FIXME# only offer the 'j' option when only one job running, and not just on CPU0
   my $multiplejob=0;
   foreach my $cpu ( sort {$multiplejob||=$ref->{job}{$a} ne $ref->{job}{$b};$a <=> $b} keys %{$ref->{job}}) {
      push(@attrs, { name => "CPU$cpu: job#", value => $ref->{job}{$cpu} });
   }
   #push(@attrs, { name => "multiple", value => $multiplejob });
   print_status_window("$nodename", \@attrs, (exists $ref->{job}{0} ? "'j'ob details on CPU0, " : "") . "'o'ffline, 'r'estore state");

}

# This is used by the 4 subs above here to actually paint the subpad
sub print_status_window {
   my $title=shift;
   my $ref=shift;
   my $epilogue=shift;

   my $line=1;
   my $string;
   my $maxlinelen;
   my $pat;
   my $indent;

   my $name;
   my $value;

   # this subwin's width is 10 fewer than the main win, and with 2 chars padding inside,
   # each line will be 14 chars fewer than the width of the main window.
   $subwin = subpad( $pad, $maxrows-4, $maxcolumns-10, 2, 5 );
   $subwin or die;
   move($subwin, $line, 0);
   clrtoeol($subwin);
   move($subwin, $line, 2);

   foreach my $attr ( @{$ref}) {
      $name=$attr->{name};
      $value=$attr->{value};

      if ($name eq "Variable_List") {
	  # strip out a messy newline character we have at ASC
	  $value =~ s/\n//;
      }

      $indent=length($name)+2+3; # 2 for padding, 3 for " = "
      $maxlinelen=($X-14)-$indent;
      $pat=".{1,$maxlinelen}";

      addstr($subwin, $name." = ");
      $string=$value;

      if ( (length($string)+$indent) > $maxlinelen) {
         while (length($string)) {

            move ($subwin, $line, $indent);
            $string=~s/($pat)// or die "can't match '$string' with '$pat'";
            addstr $subwin, "$1";
            clrtoeol($subwin);
            move($subwin, ++$line, 0);
            clrtoeol($subwin);
            move($subwin, $line, 2);

            $indent=7;
            $maxlinelen=($X-14)-$indent;
            $pat=".{1,$maxlinelen}";
         }
         move($subwin, $line, 0);
         clrtoeol($subwin);
         move($subwin, $line, 2);
      } else {
         addstr $subwin, $string;
            clrtoeol($subwin);
         move($subwin, ++$line, 0);
         clrtoeol($subwin);
         move($subwin, $line, 2);
      }
   }

   move($subwin, ++$line, 0);
   clrtoeol($subwin);
   move($subwin, $line, 2);
   if (defined $epilogue) {
       $maxlinelen=$X-14;
       $epilogue="'q' to exit, $epilogue";

       # gah, wrapping on comma is freaking hard
       if (length($epilogue) >= $maxlinelen) {
          my $subpart=substr($epilogue, 0, $maxlinelen);
          my $subremain=substr($epilogue, $maxlinelen);
          while ($subpart=~/(.*, )(.*)/) {
             addstr $subwin, $1;
             move($subwin, ++$line, 0);
             clrtoeol($subwin);
             move($subwin, $line, 2);
             my $strlen=length("$2$subremain");
             if ($strlen >= $maxlinelen) {
                $subpart=substr("$2$subremain", 0, $maxlinelen);
                $subremain=substr("$2$subremain", $maxlinelen);
             } else {
                addstr $subwin, "$2$subremain";
                last;
             }
          }
       } else {
          addstr $subwin, $epilogue;
       }
   } else {
       addstr $subwin, "'q' to exit this window";
   }
   move($subwin, ++$line, 0);
   clrtoeol($subwin);

   # make a nice box window border for our output
   resize($subwin, $line+2, $X-10);
   attron( $subwin, COLOR_PAIR( 1 )) if $colorize;
   box($subwin, &ACS_VLINE, &ACS_HLINE);
   move($subwin, 0, 3);
   addch($subwin, &ACS_RTEE);
   addstr $subwin, "$title";
   addch($subwin, &ACS_LTEE);
   attroff( $subwin, COLOR_PAIR( 1 )) if $colorize;

   $subY=$line+4;
   $subX=$X-10;  # currently not used (since the subwin is always smaller than the terminal)
   if (($subY) < $py) {
      $py=0;
   }
   pnoutrefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
}

sub update_subwin {
   $subwin or return;
   if ($searchobject{TYPE} eq "SERVER") {
        print_serverstatus_window($searchobject{VALUE});
   } elsif ($searchobject{TYPE} eq "NODE") {
        print_nodestatus_window($searchobject{VALUE}, $_[1]->{$searchobject{SERVER}}{$searchobject{VALUE}});
   } elsif ($searchobject{TYPE} eq "JOB") {
        print_jobstatus_window($searchobject{VALUE}, $searchobject{SERVER});
   } elsif ($searchobject{TYPE} eq "JOBLOAD") {
        print_jobloadstatus_window($searchobject{VALUE}, $searchobject{SERVER}, $_[1]);
   } else {
        printwarning("oddly, I'm on line ".__LINE__);
   }
}

sub destroy_subwin {
   #delwin($subwin);
   $subwin=0;
   %searchobject=();
   $subY=0;
   $subX=0;
   if ($py >= $ly - $Y + 2) {
       $py = $ly - $Y + 3;
       pnoutrefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
   }
}

sub getconfirmation {
    my $action=shift;

    my $input=getstring("Confirm $action? (type \"yes\") ");

    return ($input eq "yes");
}

#FIXME#  top_sleep() is a kludge, I know it... It just keeps growing
#FIXME#  as I add new commands.  *shrug*
sub top_sleep {

    my $targettime = time() + $sleeptime;

    while ( time() < $targettime ) {
        halfdelay(1);
        my $input = getch($cmdwin);
        if ($SIGWINCH) {
           $SIGWINCH=0;
           endwin;
           refresh();
           update_display(@_);
        }
        if ( defined $input ) {
            if ( $input eq "q" ) {
                if ($subwin) {
                    destroy_subwin();
                    update_display(@_);
                } else {
                    endwin;
                    return 1;
                }
            }
            # why doesn't curses do this automatically??
            elsif ( $input eq $CTRL_L ) {
                clear($pad);
                clrtoeol($cmdwin);
                update_display(@_);
            }
            #FIXME# $helpwin should be a scrollable pad
            elsif ( $input eq "h" || $input eq "?" ) {
                my $helpwin = newwin( 0, 0, 0, 0 );
                attron( $helpwin, A_REVERSE );
                attron( $helpwin, COLOR_PAIR(6) ) if $colorize;
                addstr $helpwin, "$script_name v$VERSION";
                attroff( $helpwin, A_REVERSE );
                attroff( $helpwin, COLOR_PAIR(6) ) if $colorize;
                addstr $helpwin, "    Backend: " . $get_version->();
                move( $helpwin, 2, 0 );
                addstr $helpwin, "Seconds Refresh ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, "$sleeptime";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nGrid Columns ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $autocolumns ? "autoset to $columns" : "$columns";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nPrimary Grid's Maximum Per-Node CPU Count ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $maxnodegrid;
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nShow CPU Number(s) ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, join ( " ", @show_cpu );
                attroff( $helpwin, A_BOLD );
                if ($show_user) {
                     addstr $helpwin, "\nLimiting job view by user to ";
                     attron( $helpwin, A_BOLD );
                     addstr $helpwin, $show_user;
                     attroff( $helpwin, A_BOLD );
                }
                if ($show_onlyq) {
                     addstr $helpwin, "\nLimiting job view by queue to ";
                     attron( $helpwin, A_BOLD );
                     addstr $helpwin, $show_onlyq;
                     attroff( $helpwin, A_BOLD );
                }
                addstr $helpwin, "\nColorization ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $colorize ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nState Summary Display ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $show_summary ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nGrid Display ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $show_grid ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nGrid Job Display ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $show_jobs ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nQueue Display ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $show_queue ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nShow Queued Jobs ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $show_qqueue ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nArray Compression ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $compress_arrays ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nNo space (compact display) ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $nospace ? "on" : "off";
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nNode sorting ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, "default => $nodesort";
                foreach (keys %nodesort_host) {
                     addstr $helpwin, "\n             $_ => ".$nodesort_host{$_};
                }
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, "\nNumber of possible colors ";
                attron( $helpwin, A_BOLD );
                addstr $helpwin, $COLOR_PAIRS;
                attroff( $helpwin, A_BOLD );
                addstr $helpwin, <<"__EOHELP__";


Interactive commands

 h       Print this help
 q       Quit
 space   Immediately Update Display

 A       Toggle Array Job Compression
 c       Grid Node Columns (0 = set to terminal width)
 C       Toggle Use of Colors in the Display
 f       Toggle Filling Background with Black
 G       Toggle Grid Display
 J       Toggle Display of Jobs in Grid
 L       Limit Job View to Specific Queue
 m       Primary grid's maximum per-node CPU count
 N       Toggle No Space (Compact Display)
 Q       Toggle Job Queue Display
 s       Seconds between display updates (e.g., 10, 2*60)
 S       Toggle State Summary Display
 t       Toggle Jobs with Status 'Q' in Job Queue Display
 u       Limit Job View to Specific User(s)

Subwindow commands ("man $script_name" for details)
 /       Search for a Server/Node/Job and Display Report in Subwindow
 n       Display Node Status Report in Subwindow
 l       Display Job Load Report in Subwindow

  Node Status Report subwindow commands    Server Status Report subwindow command
      j  Jump to Job Details Subwindow         q  Exit the Subwindow
      o  Set Node Offline
      r  Restore Node States               Job Details Report subwindow commands
      q  Exit the Subwindow                    l  Jump to Job Load Subwindow
                                               d  Delete Job
  Job Load Report subwindow commands           H  Hold Job
      j  Jump to Job Details Subwindow         R  Release Job
      q  Exit the Subwindow                    r  Re-run Job
                                               q  Exit the Subwindow
Press any key to continue...
__EOHELP__

                refresh($helpwin);
                # wait for the user to hit the any key.
                cbreak;
                nodelay( $helpwin, 0 );
                getch($helpwin);
                halfdelay(1);
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
                move( $cmdwin, 0, 0 );
                clrtoeol($cmdwin);
                refresh($cmdwin);

                delwin($helpwin);
            }
            # change the visible CPUs
            #FIXME# Don't allow the user to display CPUs that don't exist on any node
            elsif ( $input =~ /^\d$/ ) {
                if ( grep /^$input$/, @show_cpu ) {
                    @show_cpu = grep !/^$input$/, @show_cpu;
                }
                else {
                    my %seen = ();
                    foreach ( @show_cpu, $input ) {
                        $seen{$_} = 1;
                    }
                    @show_cpu = sort {$a <=> $b} (keys %seen);
                }
                update_display(@_);
            }
            elsif ( $input eq "s" ) {
                $input=getstring("Number of seconds for refresh[$sleeptime]? ");

                if ($input) {
                    my $tmp;

                    # *grin* I love this use of eval
                    if ( $tmp = eval $input and $tmp > 0 ) {
                        $sleeptime  = $tmp;
                        $targettime = time() + $sleeptime;
                    }
                    else {
                        printwarning("Invalid number!");
                    }
                }

            }
            elsif ( $input eq "c" ) {
                $input=getstring("Number of columns[$columns]? ");

                if ( $input =~ /^\d+$/ and $input >= 0 ) {
                    $columns = $input;
                    $autocolumns=!$columns;
                    update_display(@_);
                }
                elsif ($input) {
                    printwarning("Invalid number!");
                }

            }
            elsif ( $input eq "u" ) {
                $input=getstring("Limit view to \"all\", \"me\", or [+/-]username? ");
                if ($input) {
                    if ($input eq "all" or $input eq "a") {
                        $show_user=0;
                    } elsif ($input eq "me" or $input eq "m") {
                        $show_user=$ENV{USER};
                    } elsif ($input =~ /^\+(.*)/) {
                        if ($1 eq "me" or $1 eq "m") {
                           $show_user.= " ".$ENV{USER};
                        } else {
                           $show_user.= " $1";
                        }
                        $show_user =~ s/^ / /g;
                    } elsif ($input =~ /^-(.*)/) {
                        if ($1 eq "me" or $1 eq "m") {
                           $show_user =~ s/\b$ENV{USER}\b//;
                        } else {
                           $show_user =~ s/\b$1\b//;
                        }
                        $show_user =~ s/  / /g;
                        $show_user =~ s/^ / /g;
                        $show_user =~ s/ $/ /g;
                    } else {
                        $show_user=$input;
                    }
                    update_display(@_);
                }
            }
            elsif ( $input eq "L" ) {
                $input=getstring("Limit view to \"all\", or queue? ");
                if ($input) {
                    if ($input eq "all" or $input eq "a") {
                        $show_onlyq="";
                    } else {
                        $show_onlyq=$input;
                    }
                    update_display(@_);
                }
            }
            elsif ( $input eq "m" ) {
                $input=getstring("Primary grid's maximum per-node CPU count? ");
                if ($input && ($input =~ /^\d+$/) && $input > 0) {
                    $maxnodegrid=$input;
                    @show_cpu = (0 .. $maxnodegrid-1);
                    update_display(@_);
                }
            }
            elsif ( $input eq "n" ) {
                $input=getstring("Node name? ");
                
                if ($input) {
                    foreach my $server (%{ $_[1] }) {
                        if ( exists $_[1]->{$server}{$input} ) {
                             $searchobject{TYPE}="NODE";
                             $searchobject{VALUE}=$input;
                             $searchobject{SERVER}=$server;
                             $subwin=1;
                             update_display(@_);
                             last;
                        }
                    }
                }
            }
            elsif ( $input eq "j" and $subwin and $searchobject{TYPE} eq "NODE" ) {
                my $jobid=$_[1]->{$searchobject{SERVER}}{$searchobject{VALUE}}{job}{"0"};
                if (defined $jobid) {
                    $searchobject{TYPE}="JOB";
                    $searchobject{VALUE}=$jobid;
                    $searchobject{SERVER}=$_[2]->{$jobid}{server};
                    update_display(@_);
                } else {
                       printwarning("No job on CPU 0");
                }
            }
            elsif ( $input eq "j" and $subwin and $searchobject{TYPE} eq "JOBLOAD") {
                    # We are currently looking a job load report detail, immediately switch
                    # this to a job detail window
                    $searchobject{TYPE}="JOB";
                    update_display(@_);
            }
            elsif (($input eq "o" or $input eq "r") and $subwin and $searchobject{TYPE} eq "NODE") {
                my $err = $set_node_state->($searchobject{SERVER}, $searchobject{VALUE}, $input);
                if ($err) {
                    printwarning($err);
                } else {
                    $input=" ";
                }
            }
            elsif ( $input eq "d" and $subwin and $searchobject{TYPE} eq "JOB" ) {
                return unless getconfirmation("delete $searchobject{VALUE}");
                my $err = $set_job_state->($searchobject{SERVER}, $searchobject{VALUE}, 0, $input);
                if ($err) {
                    printwarning($err);
                } else {
                    $input=" ";
                }
            }
            elsif ( $input eq "r" and $subwin and $searchobject{TYPE} eq "JOB" ) {
                return unless getconfirmation("rerun $searchobject{VALUE}");
                my $err = $set_job_state->($searchobject{SERVER}, $searchobject{VALUE}, 0, $input);
                if ($err) {
                    printwarning($err);
                } else {
                    $input=" ";
                }
            }
            elsif ( $input eq "H" and $subwin and $searchobject{TYPE} eq "JOB" ) {
                return unless getconfirmation("hold $searchobject{VALUE}");
                my $err = $set_job_state->($searchobject{SERVER}, $searchobject{VALUE}, 
                        $_[2]->{$searchobject{VALUE}}{state}, $input);
                if ($err) {
                    printwarning($err);
                } else {
                    $input=" ";
                }
            }
            elsif ( $input eq "R" and $subwin and $searchobject{TYPE} eq "JOB" ) {
                return unless getconfirmation("release $searchobject{VALUE}");
                my $err = $set_job_state->($searchobject{SERVER}, $searchobject{VALUE}, 
                        $_[2]->{$searchobject{VALUE}}{state}, $input);
                if ($err) {
                    printwarning($err);
                } else {
                    $input=" ";
                }
            }
            elsif ( $input eq "l" ) {
                if ($subwin and $searchobject{TYPE} eq "JOB") {
                    # We are currently looking at a job detail, immediately switch this
                    # to a job load report
                    $searchobject{TYPE}="JOBLOAD";
                    update_display(@_);

                } else {
                    $input=getstring("Job ID Number? ");
                    if ($input) {
                        my @objects=();
                        my $searchtype="";
                        my $searchserver="";
                    
                        # what other information can we extract?
                        if ($input =~ /^\d+(?:\[\d*\])?$/ ) {
                            $searchtype||="job";
                        } elsif ($input =~ /^\d+\./ ) {
                            $searchtype||="job";
                           ($input, $searchserver)=split(/\./, $input, 2);
                        }
                        # we know everything we can, now go find stuff
                        if ($searchtype eq "job") {  
                           foreach my $job (%{ $_[2] }) {
                               if ( $input eq $job
                                    and (!$searchserver or $_[2]->{$job}{server} =~ /^$searchserver/)) {
                                   push(@objects, $job);
                                   $searchobject{TYPE}= "JOBLOAD";
                                   $searchobject{SERVER}=$_[2]->{$job}{server};
                                   $searchobject{VALUE}="$job";
                               }
                           }
                        }

                        # if we have anything useful, go display it
                        if (scalar @objects > 1) {
                            printwarning("Multiple objects found.  Please narrow your search.");
                        } elsif (scalar @objects < 1) {
                            printwarning("no objects found");
                        } elsif (exists $searchobject{TYPE} && defined $searchobject{TYPE}) {
                            $subwin=1;
                            update_display(@_);
                        }
                    }
                }
            }
            # Just about all of this should be moved out of here
            #FIXME# need to unify these searches somehow
            elsif ( $input eq "/" ) {
                $input=getstring("Search string? ");
            
                if ($input) {
                    my @objects=();
                    my $searchtype="";
                    my $searchserver="";

                    # did the user specify a pattern?
                    if ($input =~ s/^~(.)\s?//) {
                        if ($1 eq "s") {
                            $searchtype="server";
                        } elsif ($1 eq "j") {
                            $searchtype="job";
                        } elsif ($1 eq "n") {
                            $searchtype="node";
                        } else {
                            printwarning("Invalid search pattern");
                            next;
                        }
                    }
                    # what other information can we extract?
                    if ($input =~ /^\d+(?:\[\d*\])?$/ ) {
                        $searchtype||="job";
                    } elsif ($input =~ /^\d+-\d+/ ) {
                        $searchtype||="job";
                       ($input, $searchserver)=split(/\./, $input, 2);
                    } elsif ($input =~ /^\d+(?:\[\d*\])?\./ ) {
                        $searchtype||="job";
                       ($input, $searchserver)=split(/\./, $input, 2);
                    } elsif ($input =~ /^[a-z]\w*\s+\w/) {
                        $searchtype||="node";
                       ($input, $searchserver) = split(/\s+/,$input, 2);
                    }
                        
                    # we know everything we can, now go find stuff
                    if ($searchtype eq "job") {  
                       foreach my $job (%{ $_[2] }) {
                          if ( $input eq $job
                               and (!$searchserver or $_[2]->{$job}{server} =~ /^$searchserver/i)) {
                              push(@objects, $job);
                              $searchobject{TYPE}="JOB";
                              $searchobject{SERVER}=$_[2]->{$job}{server};
                              $searchobject{VALUE}="$job";
                          }
                       }
                    }

                    if ($searchtype eq "node" or !$searchtype) {
                       if ($searchserver) {
                          if (exists $_[1]->{$searchserver} and exists $_[1]->{$searchserver}{$input}) {
                             push(@objects, $input);
                             $searchobject{TYPE}="NODE";
                             $searchobject{SERVER}="$searchserver";
                             $searchobject{VALUE}="$input";
                          } else {
                             foreach my $server (%{ $_[1] }) {
                                if ($server =~ /$searchserver/i) {
                                   if (exists $_[1]->{$server} and exists $_[1]->{$server}{$input}) {
                                      push(@objects, $input);
                                      $searchobject{TYPE}="NODE";
                                      $searchobject{SERVER}="$server";
                                      $searchobject{VALUE}="$input";
                                   }
                                }
                             }
                          }
                             
                       } else  {
                          foreach my $server (%{ $_[1] }) {
                             if (exists $_[1]->{$server} and exists $_[1]->{$server}{$input} ) {
                                push(@objects, $input);
                                $searchobject{TYPE}="NODE";
                                $searchobject{SERVER}="$server";
                                $searchobject{VALUE}="$input";
                             }
                          }
                       }
                    }
                          
                    if ($searchtype eq "server" or !$searchtype) {
                       foreach my $server (%{ $_[1] }) {
                          if ($server =~ /$input/i) {
                             push(@objects, $input);
                             $searchobject{TYPE}="SERVER";
                             $searchobject{SERVER}="$server";
                             $searchobject{VALUE}="$server";
                          }
                       }
                    }

                    # if we have anything useful, go display it
                    if (scalar @objects > 1) {
                        printwarning("Multiple objects found.  Please narrow your search.");
                    } elsif (scalar @objects < 1) {
                        printwarning("no objects found matching $input ($searchtype)");
                    } elsif (exists $searchobject{TYPE} && defined $searchobject{TYPE}) {
                        $subwin=1;
                        update_display(@_);
                    }
                }
            }
            # all of these toggles should be self-explanatory
            elsif ( $input eq "C" ) {
                $colorize = !$colorize;
                if ($colorize && !has_colors() ) {
                    printwarning("Terminal doesn't support colors");
                    $colorize=0;
                }
                update_display(@_);
            }
            elsif ( $input eq "f" ) {
                $fillbg = !$fillbg;
                init_pairs();
                update_display(@_);
            }
            elsif ( $input eq "N" ) {
                $nospace = !$nospace;
                update_display(@_);
            }
            elsif ( $input eq "G" ) {
                $show_grid = !$show_grid;
                update_display(@_);
            }
            elsif ( $input eq "S" ) {
                $show_summary = !$show_summary;
                update_display(@_);
            }
            elsif ( $input eq "Q" ) {
                $show_queue = !$show_queue;
                update_display(@_);
            }
            elsif ( $input eq "t" ) {
                $show_qqueue = !$show_qqueue;
                update_display(@_);
            }
            elsif ( $input eq "J" ) {
                $show_jobs = !$show_jobs;
                update_display(@_);
            }
            elsif ( $input eq "A" ) {
                $compress_arrays = !$compress_arrays;
                $reletterize = 1;
                letterize( \%Jobs );
                colorize( \%Jobs );
                $reletterize = 0;
                update_display(@_);
            }
            #FIXME# my home keyboard sends FIND and SELECT instead of HOME and END, weird?
            elsif ( $input eq KEY_HOME or $input eq KEY_SHOME or $input eq KEY_FIND ) {
                $py = 0;
                $px = 0;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq KEY_END or $input eq KEY_SEND  or $input eq KEY_SELECT) {
                $py = ($ly>$subY?$ly:$subY) + 2 - $Y;
                $px = 0;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq KEY_PPAGE or $input eq KEY_SPREVIOUS 
                    or $input eq $CTRL_B) {
                $py -= $Y -2;
                $py <= 0 and $py = 0;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq KEY_NPAGE or $input eq KEY_SNEXT 
                    or $input eq $CTRL_F) {
                $py += $Y - 2;
                $py >= ($ly>$subY?$ly:$subY) - $Y + 2 and $py = ($ly>$subY?$ly:$subY) + 2 - $Y;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq "k" or $input eq KEY_UP ) {
                $py <= 0 and $py = 0, next;
                $py--;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq "j" or $input eq KEY_DOWN) {
                $py >= ($ly>$subY?$ly:$subY) - $Y + 2 and next;
                $py++;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq "h" or $input eq KEY_LEFT ) {
                $px <= 0 and $px = 0, next;
                $px -= 2;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            elsif ( $input eq "l" or $input eq KEY_RIGHT ) {
                $px >= $lx - $X + 1 and next;
                $px += 2;
                prefresh( $pad, $py, $px, 0, 0, $Y - 2, $X - 1 );
            }
            if ( $input eq " " ) {
                addstr $cmdwin, 0, 0, "Updating...";
                clrtoeol($cmdwin);
                refresh($cmdwin);
                move( $cmdwin, 0, 0 );
                clrtoeol($cmdwin);
                return 0;
            }
            # for debugging
            #else {
                #addstr $cmdwin, 0, 0, ord($input);
            #}
        }
    }
}

sub readrc {
   my ($f)=shift or return;
   return unless (-f $f);

   open(F, $f) or die "$f: $!\n";
   while ($_=<F>) {
      chomp;
      s/#.*//;
      next unless $_ =~ /=/;
      my ($name, $value) = split (/=/,$_,2);
      $name =~ s/\s//g;
      $value =~ s/\s//g;
      next if (length($name) <=0 or length($value) <=0);

      if ($name eq "host" or $name eq "show_cpu" ) {
         # /em cringes
         eval "\@$name=(\"".join('","', split(',', $value))."\")";
         if (scalar @show_cpu == 1 and $show_cpu[0] eq '()') {
            @show_cpu = ();
         }
      } elsif ($name eq "show_user" ) {
         eval "\$$name=join(' ', split(',', \"$value\"))";
      } elsif ($name eq "nodesort_host" ) {
         @nodesort_hostconf=(makehashfromrc($value));
      } else {
         eval "\$$name=$value";
      }
   }
   close F;
}

sub makehashfromrc {
   my ($value)=@_;
   my @tmplist=();
   $value =~ s/^\(//;
   $value =~ s/\)$//;
   foreach my $valpair (split(",",$value)) {
       unshift(@tmplist, [split("=",$valpair)]);
   }
   return @tmplist;
}

sub init_nodesort {
   my (@hosts)=@_;

   # %nodesort_host needs to be initialized

   foreach my $host (@hosts) {

      # first check for exact matches
      foreach my $conf (@nodesort_hostconf) {
         if ($host eq $conf->[0]) {
            $nodesort_host{$host}=$conf->[1];
         }
      }
      # then check for regex matches
      if (!exists $nodesort_host{$host}) {
         foreach my $conf (@nodesort_hostconf) {
            if ($host =~ /$conf->[0]/) {
               $nodesort_host{$host}=$conf->[1];
            }
         }
      }
      # then just fill in the default
      if (!exists $nodesort_host{$host}) {
         $nodesort_host{$host}=$nodesort;
      }
   }
}


## node sorting functions
sub node_sort_lexical { sort keys %{$_[0]}; }

sub node_sort_mixed {
 map { $_->[0] }
     sort { $b->[1] cmp $a->[1]
                    ||
            $a->[2] <=> $b->[2]
                    ||
            $b->[0] cmp $a->[0]
          } map { [$_, /([[:alpha:]]+)/, /([[:digit:]]+)/] } keys %{$_[0]};
}

sub node_sort_mixed2 {
 map { $_->[0] }
     sort { $a->[1] cmp $b->[1]
                    ||
            $a->[2] <=> $b->[2]
                    ||
            $a->[3] cmp $b->[3]
                    ||
            $a->[4] <=> $b->[4]
                    ||
            $b->[0] cmp $b->[0]
          } map { /^([[:alpha:]]+).*([[:digit:]]+).*([[:alpha:]]+).*([[:digit:]]+)/; [$_, $1, $2, $3, $4] } keys %{$_[0]};
}

sub node_sort_integer {
 map { $_->[0] }
     sort { $a->[1] <=> $b->[1]
                    ||
            $b->[0] cmp $b->[0]
          } map { [$_, /([[:digit:]]+)/] } keys %{$_[0]};
}

sub node_sort_ordered {
  sort { $_[0]->{$a}{rank} <=> $_[0]->{$b}{rank} } keys %{$_[0]};
}

sub node_sort_func {
  my ($host, $nodehash)=@_;

  unless (exists $nodesort_host{$host} and defined $nodesort_host{$host}) {
    printwarning("unknown host ($host, $nodehash)");
    sleep 2;
  }

  my $enodesort=$nodesort_host{$host};

  if ($enodesort eq "integer") {
      return node_sort_integer($nodehash);
  } elsif ($enodesort eq "mixed") {
      return node_sort_mixed($nodehash);
  } elsif ($enodesort eq "mixed2") {
      return node_sort_mixed2($nodehash);
  } elsif ($enodesort eq "lexical") {
      return node_sort_lexical($nodehash);
  } elsif ($enodesort eq "ordered") {
      return node_sort_ordered($nodehash);
  }
  printwarning("unknown nodesort ($host, $nodehash)");
  sleep 2;
  return node_sort_lexical($nodehash);

}

sub count_nodestring {
  my $nodes=shift;
  my $nodect=0;

  foreach my $req (split(/\+/,$nodes)) {
    if ($req =~ m/(\d+):.*ppn=(\d+)/) {
      $nodect += $1 * $2;
    } elsif ($req =~ m/^(\d+)(:.*)?$/) {
      $nodect +=  $1;
    }
  }
  return $nodect;
}

sub init_pairs {
   my $fillbgCol;
   if ($fillbg) {
      $fillbgCol=COLOR_BLACK;
      assume_default_colors(COLOR_WHITE,COLOR_BLACK);
   } else {
      $fillbgCol=-1;
      assume_default_colors(-1,-1);
   }
   # This is every possible color combo list below.  Over time, I've commented out
   # color pairs that don't look very good.  If your eyes disagree with my eyes,
   # you are free to play around with this list.  But don't forget... only the
   # first $COLOR_PAIRS uncommented combos apply.  $COLOR_PAIRS is set by your
   # curses implementation.  %SCRIPTNAME%'s help screen (hit 'h' in %SCRIPTNAME%) will tell
   # you the value of $COLOR_PAIRS.
   init_pair( ++$pr, COLOR_RED,      $fillbgCol );
   init_pair( ++$pr, COLOR_GREEN,    $fillbgCol );
   init_pair( ++$pr, COLOR_YELLOW,   $fillbgCol );
   init_pair( ++$pr, COLOR_BLUE,     $fillbgCol );
   init_pair( ++$pr, COLOR_MAGENTA,  $fillbgCol );
   init_pair( ++$pr, COLOR_CYAN,     $fillbgCol );
   #init_pair( ++$pr, COLOR_WHITE,    $fillbgCol );
   #init_pair( ++$pr, COLOR_BLACK,   COLOR_BLACK );
   #init_pair( ++$pr, COLOR_RED,     COLOR_WHITE );
   #init_pair( ++$pr, COLOR_GREEN,   COLOR_WHITE );
   #init_pair( ++$pr, COLOR_YELLOW,  COLOR_WHITE );
   #init_pair( ++$pr, COLOR_BLUE,    COLOR_WHITE );
   #init_pair( ++$pr, COLOR_MAGENTA, COLOR_WHITE );
   #init_pair( ++$pr, COLOR_CYAN,    COLOR_WHITE );
   #init_pair( ++$pr, COLOR_WHITE,   COLOR_WHITE );
   init_pair( ++$pr, COLOR_BLACK,   COLOR_WHITE );
   #init_pair( ++$pr, COLOR_RED,     COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_GREEN,   COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_YELLOW,  COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_BLUE,    COLOR_YELLOW );
   init_pair( ++$pr, COLOR_MAGENTA, COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_CYAN,    COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_WHITE,   COLOR_YELLOW );
   #init_pair( ++$pr, COLOR_BLACK,   COLOR_YELLOW );
   init_pair( ++$pr, COLOR_RED,     COLOR_CYAN );
   #init_pair( ++$pr, COLOR_GREEN,   COLOR_CYAN );
   init_pair( ++$pr, COLOR_YELLOW,  COLOR_CYAN );
   #init_pair( ++$pr, COLOR_BLUE,    COLOR_CYAN );
   init_pair( ++$pr, COLOR_MAGENTA, COLOR_CYAN );
   #init_pair( ++$pr, COLOR_CYAN,    COLOR_CYAN );
   init_pair( ++$pr, COLOR_WHITE,   COLOR_CYAN );
   init_pair( ++$pr, COLOR_BLACK,   COLOR_CYAN );
   init_pair( ++$pr, COLOR_RED,     COLOR_MAGENTA ); 
   init_pair( ++$pr, COLOR_GREEN,   COLOR_MAGENTA );
   init_pair( ++$pr, COLOR_YELLOW,  COLOR_MAGENTA );  # current 16th
   init_pair( ++$pr, COLOR_BLUE,    COLOR_MAGENTA );
   #init_pair( ++$pr, COLOR_MAGENTA, COLOR_MAGENTA );
   init_pair( ++$pr, COLOR_CYAN,    COLOR_MAGENTA );
   init_pair( ++$pr, COLOR_WHITE,   COLOR_MAGENTA );
   #init_pair( ++$pr, COLOR_BLACK,   COLOR_MAGENTA );
   #init_pair( ++$pr, COLOR_RED,     COLOR_RED );
   init_pair( ++$pr, COLOR_GREEN,   COLOR_RED );
   init_pair( ++$pr, COLOR_YELLOW,  COLOR_RED );
   init_pair( ++$pr, COLOR_BLUE,    COLOR_RED );
   init_pair( ++$pr, COLOR_MAGENTA, COLOR_RED );
   init_pair( ++$pr, COLOR_CYAN,    COLOR_RED );
   init_pair( ++$pr, COLOR_WHITE,   COLOR_RED );
   init_pair( ++$pr, COLOR_BLACK,   COLOR_RED );
   init_pair( ++$pr, COLOR_RED,     COLOR_GREEN );
   #init_pair( ++$pr, COLOR_GREEN,   COLOR_GREEN );
   #init_pair( ++$pr, COLOR_YELLOW,  COLOR_GREEN );
   init_pair( ++$pr, COLOR_BLUE,    COLOR_GREEN );
   #init_pair( ++$pr, COLOR_MAGENTA, COLOR_GREEN );
   init_pair( ++$pr, COLOR_CYAN,    COLOR_GREEN );
   init_pair( ++$pr, COLOR_WHITE,   COLOR_GREEN );
   init_pair( ++$pr, COLOR_BLACK,   COLOR_GREEN );
}

sub init_colors {
    return ( 1 .. ($COLOR_PAIRS-1 > $pr ? $pr : $COLOR_PAIRS-1) );
}

1;

__END__
